 /*
    * Copyright (c) 2012, H. J. Ohlbach. All rights reserved.
    * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
    *
    * This code is free software; you can redistribute it and/or modify it
    * under the terms of the GNU General Public License version 2 only, as
    * published by the Free Software Foundation. 
    *
    * This code is distributed in the hope that it will be useful, but WITHOUT
    * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    * version 2 for more details.
    */
package Java2OWL;

import java.util.ArrayList;
import java.util.Set;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLDataProperty;
import org.semanticweb.owlapi.model.OWLNamedIndividual;
import org.semanticweb.owlapi.model.OWLObjectProperty;

/** The instances of this class maintain the correspondences between Java-objects and OWL-individuals.
 *  <p>
 *  If OWL-individuals are created from Java-objects then there is a one-to-one correspondence between 
 *  the two. If, however, Java-objects are created from OWL-individuals then, 
 *  due to multi-inheritance of OWL and mono-inheritance of Java, a particular
 *  OWL-individual may be mapped to several Java-objects. 
 *  An example is the Amphibious-vehicle as subclass of ship and surface vehicle. 
 *  It would be mapped to both a Java Ship and a Java SurfaceVehicle. 
 * 
 * <p>All methods are thread safe. The monitor is the individual wrapper itself. 
 *
 * @author ohlbach
 */
public class IndividualWrapper {
    
    /** the top manager.*/
    private J2OManager manager;
    
    /** the list of Java-objects. */
    private Object[] objects; 
    
    /** the OWL-individual. */
    private OWLNamedIndividual individual;
    
    /** indicates the position of the synchronisation flag within J2OSynchroniser.SYNCHRONIZE_BYOBJECT_VARIABLE
     * <p> This variable is part of the synchronisation code injected into the annotated classes.
     * The synchronisation flag controls activation of synchronisation at object level.
     */
    static final int synchronisationFlag = 1;
    
    /** indicates the position of the changed since synchronisation flag within J2OSynchroniser.SYNCHRONIZE_BYOBJECT_VARIABLE
     * <p> This variable is part of the synchronisation code injected into the annotated classes.
     * The changed since synchronisation flag prevents unnecessary synchronisation. 
     */
    private static final int changedSinceSynchronisationFlag = 2;
   
    /** indicates the position of the changed since reclassification flag within J2OSynchroniser.SYNCHRONIZE_BYOBJECT_VARIABLE
     * <p> This variable is part of the synchronisation code injected into the annotated classes.
     * The changed since reclassification flag prevents unnecessary reclassification. 
     */
    private static final int changedSinceReclassificationFlag = 4;
    
    /** a constructor with both, a Java-object and an OWL-individual. 
     * 
     * @param object        the Java-object.
     * @param individual    the OWL-individual.
     * @param individualManager a link to the corresponding J2OIndividualManager.
     */
    IndividualWrapper(Object[] objects, OWLNamedIndividual individual, J2OManager manager) {
        this.objects = objects;
        this.individual = individual;
        this.manager = manager;}
    
 
    /** maps an OWL-individual to corresponding Java-objects. 
      * 
      * @param individual an OWLNamedIndividual
      * @return the corresponding IndividualWrapper or null, if, for example the OWL-classes are not mapped to Java classes.
      * @throws J2OException if something goes wrong.
      * 
      * <p>Since OWL allows for multiple inheritance, and Java does not, the system may map a single
      * OWL-individual to several Java-objects. 
      * As an example an instance of a amphibian-vehicle (a subclass of Ship and Surface-Vehicle) would
      * be mapped to both, a ship and a surface-vehicle.
      * The correspondences are kept in the IndividualWrapper structure.
      */
    IndividualWrapper (OWLNamedIndividual individual, J2OManager manager) throws J2OException {
        this.manager = manager;
        this.individual = individual;
        J2OOntologyManager j2oOntologyManager = manager.getOntologyManager();
        Set<OWLClass> ocls = j2oOntologyManager.getOWLClasses(individual);
        if(ocls == null) throw new J2OException("Reasoner Error");
        J2OClassManager j2oClassManager = manager.getClassManager();
        Set<ClassWrapper> clwrps = j2oClassManager.getCLWs(ocls);
        if(clwrps.isEmpty()) {
            String s = "";
            for(OWLClass ocl : ocls) s += ocl.toStringID() + ", ";
            throw new J2OException("OWL classes containing to the OWL individual " + 
                    individual.toStringID() + " are not mapped to Java: " + s);}
        objects = new Object[clwrps.size()];
        if(!ClassWrapper.instantiate(clwrps, objects,0,manager.getLogger()))
            throw new J2OException("Instantiation Error");
        J2OIndividualManager individualManager = manager.getIndvidualManager();
        
        individualManager.insertIWR(this);
        // This is necessary to avoid infinite loops caused be loops in the role graph.
        // The loops may show up in transferOWLProperties.
        for(int i = 0; i < objects.length; ++i) {
            if(!transferProperties(individual,objects,i,false)) {
                individualManager.removeIWR(this);
                throw new J2OException("Transfer Error");}}}
    
    
   /** transfers all the common fields of objects[0]...object[i-1] to object[i] and takes the remaining fields from the individual.
     * 
     * @param individual an OWL individual
     * @param objects the list of Java objects to which the individual is mapped.
     * @param i the index of the object to be filled with field values.
     * @return true if the operation succeeded.
     */ 
    private boolean transferProperties(OWLNamedIndividual individual, Object[] objects, int i, boolean forceReclassification) {
        Object object = objects[i];
        Class jcl = object.getClass();
        if(i > 0) Utilities.transferCommonFields(objects,i,object,jcl);
        return manager.getIndvidualManager().transferOWLProperties(individual,object,jcl,forceReclassification);}


    /** returns the individual. */
    public OWLNamedIndividual getIndividual() {return individual;}
    
    /** returns all the Java objects. */
    synchronized public Object[] getObjects() {return objects;}
    
    
    /** returns the first object in the stored objects which is an instance of the given class.
     * 
     * @param jcl a Java class
     * @return the first object which is an instance of the given class or null, if none could be found,
     */
    synchronized public Object getObject(Class jcl){
        for(Object object : objects) {
            if(jcl.isInstance(object)) return object;}
        return null;} 
     
    /** returns the first Java object.
     * No check is performed to ensure that there is a first object at all.
     */
    synchronized public Object getFirstObject() {return objects[0];}
    
    /** sets a data property value for a data property which has no corresponding field in the objects. 
     * <p>The value is forwarded to the ontology.
     * @return true if the operation succeeded.*/
    synchronized public boolean setDataPropertyValue(OWLDataProperty property, Object value) {
        return manager.getOntologyManager().setDataProperty(individual,property,value);}
    
    /** sets an object property value for an object property which has no corresponding field in the objects.
     * <p>The value is forwarded to the ontology.
     * @return true if the operation succeeded.*/
    synchronized public boolean setObjectPropertyValue(OWLObjectProperty property, Object value) {
        return manager.getIndvidualManager().setObjectPropertyValue(individual,property,value);}
    
    /** gets a data property value for a data property which has no corresponding field in the objects, or null if an error occurred. 
     * <p>The value is forwarded to the ontology.
     * @return the set of data values as Java objects (maybe wrapper objects for numbers).
     */
    synchronized public Set<Object> getDataPropertyValues(OWLDataProperty property) {
        return manager.getOntologyManager().getDataPropertyValues(individual,property);}
   
    /** gets object properties value for an object property which has no corresponding field in the objects. 
     * 
     * @param property an OWLObjectProperty
     * @return an array of IndividualWrappers for the property's role fillers.
     */
    synchronized public IndividualWrapper[] getObjectPropertyValues(OWLObjectProperty property) {
        return manager.getIndvidualManager().getObjectPropertyValues(individual,property);}
   
    
    /** puts the value into all object's fields with the given name 
     * 
     * @param name the name of a field of some of the objects.
     * @param value the object to be put into the field.
     * <p>Objects without such a field or if the field is not accessible, are ignored.
     * <p>This method can be used to put pointer-equal values into the corresponding fields of all objects.
     */
    synchronized public void setFieldValue(String name, Object value) {
        for(Object object : objects) {
            try{Utilities.setObjectField(object, name, value);}
            catch(Exception ex) {}}}
    
    
    /** returns the objects as a |-separated string. 
     * <p> The method uses the toString() method of the objects.*/
    synchronized public String getObjectsString() {
        String s = ""; 
        for(int i = 0; i < objects.length-1; ++i) s += objects[i].toString() + " | ";
        if(objects.length >= 0) s +=  objects[objects.length-1].toString();
        return s;}
    
    
     /** activates Java <-> OWL synchronisation for all objects in the individual wrapper.
     * 
     * @param log if 'true' then the activation is logged.
     * <p> All previous changes which have not yet been synchronised get synchronised. 
     */
    synchronized public void activateSynchronisation(boolean log) {
        synchroniseWithOWL();
        for(Object object : objects) {
            if(log)manager.getLogger().info("synchronisation activated for object", object.getClass(), object.hashCode());
            try{Utilities.resetObjectFlag(object, J2OSynchroniser.SYNCHRONIZE_BYOBJECT_VARIABLE, synchronisationFlag);}
            catch(Exception ex){
                manager.getLogger().error("Field access error", Logging.Origin.JAVA, object.getClass(),J2OSynchroniser.SYNCHRONIZE_BYOBJECT_VARIABLE, ex);}}}
    
    
    
    /** activates Java <-> OWL  synchronisation for all object in the individual wrapper.
     * 
     * @param log if 'true' then the deactivation is logged.
     */
    synchronized public void deactivateSynchronisation(boolean log) {
        for(Object object : objects) {
            if(log)manager.getLogger().info("synchronisation deactivated for object", object.getClass(), object.hashCode());
            try{Utilities.setObjectFlag(object, J2OSynchroniser.SYNCHRONIZE_BYOBJECT_VARIABLE, synchronisationFlag);}
            catch(Exception ex){
                manager.getLogger().error("Field access error", Logging.Origin.JAVA, object.getClass(),J2OSynchroniser.SYNCHRONIZE_BYOBJECT_VARIABLE, ex);}}}
    
    /** synchronises the attributes of all objects with the properties of the individual.*/
    synchronized public void synchroniseWithOWL() {
        J2OOntologyManager omg = manager.getOntologyManager();
        for(Object object : objects) {
            try {
                if(Utilities.getAndResetObjectFlag(object, J2OSynchroniser.SYNCHRONIZE_BYOBJECT_VARIABLE, changedSinceSynchronisationFlag))
                    omg.synchroniseWithOWL(object, individual);}
            catch(Exception ex){// if the class contains no sychroniser code, we do block synchronisation anyway. 
                omg.synchroniseWithOWL(object, individual);}}}
    
    /** This method can reclassify the Java objects if the changed attributes puts them into another OWL-concept. 
     * 
     * <p>Example: Consider a class Student with attribute semester.<br/>
     * Consider a subclass Freshman which are Students in their first semester. <br/>
     * Consider a student Paul in the second semester, which is put back into the first semester. 
     * Now the instance Paul of the class Student should become an instance Paul of the class Freshman. 
     * <p>In detail: <br/>
     * First the OWLIndividual is realised by the OWL-reasoner, 
     * thus getting the lowest classes in the OWL hierarchy. 
     * Then it searches the OWL classes upwards to find the lowest classes which have corresponding
     * Java classes.
     * For all these classes C:
     * <ul>
     * <li>If an old object has the same class or a more special class then this object is untouched. <br/>
     *     An object which became generalised need not be reclassified.
     * </li>
     * <li>If an old object is of a superclass of C, i.e. the individual got more specialised,
     * then a new object is created and the fields if the old objects are transferred to the new object.</li>
     * <li>For all remaining classes new objects are created and their field-contents is taken 
     * from the individual's role values. 
     * </ul>
     * <p>No change to the object is done if something goes wrong.
     * 
     * <p>The reclassification has a nondeterminism in the algorithm: <br/>
     * Suppose the individual I has an object property P with filler F (an OWL individual)
     * Suppose F is mapped to two or more Java Objects o1,o2,...
     * If property P has a setter S with argument type T, then it is not clear which of the
     * objects o1,o2,... to be filled into S. <br/>
     * The algorithm chooses the first one whose type is T or a subtype of T. <br/>
     * If none of the objects has an appropriate type then an error is generated.
     * 
     * @param forceReclassification if this flag is false then reclassification is done only if some 
     * attributes have been changed since the last reclassification.
     * 
     * @return true if the operation succeeded.
     * 
     */
    synchronized public boolean reclassify(boolean forceReclassification) {
        J2OIndividualManager individualManager = manager.getIndvidualManager();
        individualManager.clearPropertyValueCache();
        boolean doReclassification = forceReclassification;
        
        for(Object object : objects) {
            try{doReclassification |= Utilities.getAndResetObjectFlag(object, J2OSynchroniser.SYNCHRONIZE_BYOBJECT_VARIABLE, changedSinceReclassificationFlag);}
            catch(Exception ex){doReclassification = true;}} // the class might not have synchroiser code.
        
        if(doReclassification){
            J2OOntologyManager j2oOntologyManager = manager.getOntologyManager();
            J2OClassManager j2oClassManager = manager.getClassManager();
            Logging logger = manager.getLogger();
            Set<OWLClass> newOWLClasses = j2oOntologyManager.getOWLClasses(individual);
            if(newOWLClasses == null) return false; // reasoner caused an error.
            Set<ClassWrapper> newCLWs = j2oClassManager.getCLWs(newOWLClasses);
            if(newCLWs.isEmpty()) {
                logger.error("Reclassification: OWL classes are not mapped to Java", Logging.Origin.USAGE, individual, newOWLClasses);
                return false;}
            Object[] oldObjects = objects;
      
            ArrayList<Object> newObjects = new ArrayList<Object>();
            Object newObject = null;
        
            // At first we test whether the old objects can be kept as they are or whether they became more specific.
        LOOP: 
            for(Object object : oldObjects) { 
                newObject = null;
                Class oldClass = object.getClass();
                for(ClassWrapper newCLW : newCLWs) {
                    Class newClass = newCLW.getJavaClass(); 
                    
                    // if we have already a new object that does the job, we don't need another one.
                    for(Object no : newObjects) {
                        if(newClass.isAssignableFrom(no.getClass())) { 
                            newCLWs.remove(newCLW);}
                            continue LOOP;}
                    
                    // the old object is more special, we can keep it as it is.
                    if(newClass.isAssignableFrom(oldClass)) { 
                        newCLWs.remove(newCLW);
                        newObject = object; 
                        break;}
                    
                    // The new class is more special, we need a new object. 
                    if(oldClass.isAssignableFrom(newClass)) {
                        newCLWs.remove(newCLW);
                        try {
                            newObject = newClass.newInstance(); 
                            Utilities.transferFields(object, newObject,newClass);
                            break;} 
                        catch (Exception ex) {
                            logger.error("Creating new object failed.",Logging.Origin.JAVA,object,ex);
                            return false;}}}
                if(newObject != null) {newObjects.add(newObject);}}
     
            // The remaining new class wrappers denote completely different classes.
            Object[] reclassifiedObjects = new Object[newObjects.size()+newCLWs.size()];
            int i = 0;
            for(Object object : newObjects) {reclassifiedObjects[i] = object; ++i;}
            if(!ClassWrapper.instantiate(newCLWs,reclassifiedObjects,i,logger)) {return false;}
        
            individualManager.removeIWR(this);
            objects = reclassifiedObjects;
            individualManager.insertIWR(this);
            //This avoids infinite loops caused by loops in the role graph of the individual.
        
            for(int j = 0; j < reclassifiedObjects.length; ++j) {
                if(!transferProperties(individual,reclassifiedObjects,j,forceReclassification)) {
                    individualManager.removeIWR(this);
                    objects = oldObjects; // restore the old state.
                    individualManager.insertIWR(this);
                    return false;}}}
        return true;}
    
    /** returns a string of the kind 'individual name <-> object name_1, ..., object name_n'*/
    public String getName() {
        String s = individual.getIRI().getFragment() + "<->";
        for(Object object : objects) {s += Utilities.getName(object) + ", ";}
        return s.substring(0,s.length()-2);}
    
    
    /** returns a string representation of the individual wrapper.*/
    @Override
    public String toString() {
        String s = individual.toString() + " <-> ";
        for(Object object : objects) {s += "\n"+object.toString();}
        return s + "\n Manager: " + manager.getName();}

}
