/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package Java2OWL;

import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.*;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.semanticweb.owlapi.expression.ParserException;
import org.semanticweb.owlapi.model.*;

/** This class is for mapping Java-Objects to OWL-individuals and vice versa.
 * <p>
 *  It provides the following services:
 * <ol>
 *  <li>One can insert new Java-objects and create corresponding OWL-individuals</li>
 *  <li>If an attribute of a Java-object is changed, one can change the corresponding attribute of the OWL-individual.</li>
 *  <li>One can map OWL-individuals to Java-objects.</li>
 * </ol>
 * <p>
 *     Since OWL allows for multiple inheritance, and Java does not, the system may map a single
 *     OWL-individual to several Java-objects. 
 *     As an example an instance of a amphibian-vehicle (a subclass of Ship and Surface-Vehicle) would
 *     be mapped to both, a ship and a surface-vehicle.
 *     The correspondences are kept in the IndividualWrapper structure.
 * 
 * @author ohlbach
 */
public class J2OIndividualManager {
    /** the top manager. */ 
    private J2OManager manager; 
    
    /** returns the top manager. */
    public J2OManager getJ2OManager() {return manager;} 
    
    /** returns the name of the individual manager.*/
    public String getName() {return manager.getName() + "_IndividualManager";}
    
    /** the J2OClassManager */
    private J2OClassManager j2oClassManager;         
    
    /** the J2OOntologyManager */
    private J2OOntologyManager j2oOntologyManager = null;   
    
    /** the logger */
    private Logging logger = null;
    
    /** maps Java-objects to IndividualWrappers. */
    private final Map<Object,IndividualWrapper> obj2iwr= new HashMap<Object,IndividualWrapper>();
    

    /** maps OWL-individuals to IndividualWrappers. */
    private final Map<OWLNamedIndividual,IndividualWrapper> ind2iwr= new HashMap<OWLNamedIndividual,IndividualWrapper>();
    
    /** maps Java classes to the individual wrappers with the objects contained in these classes. */
    private final Map<Class,HashSet<IndividualWrapper>> class2iwr = new HashMap<Class,HashSet<IndividualWrapper>>();
    
    /** inserts the individual wrapper into the internal lists. */
    void insertIWR(IndividualWrapper iwr) {
        Object[] objects = iwr.getObjects();
        synchronized(obj2iwr) {for(Object object : objects) obj2iwr.put(object, iwr);}
        synchronized(ind2iwr) {ind2iwr.put(iwr.getIndividual(), iwr);}
        synchronized(class2iwr) {
            for(Object object : objects) {
                Class jcl = object.getClass();
                HashSet<IndividualWrapper> iwrps = class2iwr.get(jcl);
                if(iwrps == null) {
                    iwrps = new HashSet<IndividualWrapper>();
                    class2iwr.put(jcl,iwrps);}
                iwrps.add(iwr);}}}
    
     /** removes the individual wrapper from the internal lists */
     void removeIWR(IndividualWrapper iwr) {
        Object[] objects = iwr.getObjects();
        synchronized(obj2iwr) {for(Object object : objects) obj2iwr.remove(object);}
        synchronized(ind2iwr) {ind2iwr.remove(iwr.getIndividual());}
        synchronized(class2iwr) {
            for(Object object : objects) {
                Class jcl = object.getClass();
                HashSet<IndividualWrapper> iwrps = class2iwr.get(jcl);
                if(iwrps != null) {iwrps.remove(iwr);}}}}
   
    /** return the individual wrappers stored for the given class, or null if there are none.
      * 
      * @param jcl a Java class
      * @return the individual wrappers stored for the given class or the lowest superclass of jcl, 
      * or null if there are none at all.
      */
    public Set<IndividualWrapper> getIWRs(Class jcl) {
        Set<IndividualWrapper> result = null;
        while(jcl != null) {
            if((result = class2iwr.get(jcl)) != null) return result.isEmpty() ? null : result;
            jcl = jcl.getSuperclass();}
        return null;}
   
    
    /** @return the individual wrapper for the object, or null if there is none. */
    public IndividualWrapper getIWR(Object object) {
        synchronized(obj2iwr) {return obj2iwr.get(object);}}
    
    /** @return the individual wrapper for the individual, or null if there is none. */
    public IndividualWrapper getIWR(OWLNamedIndividual individual) {
        synchronized(ind2iwr) {return ind2iwr.get(individual);}}
    
    /** creates an individual manager and imports all the other managers */
    J2OIndividualManager(J2OManager manager) {
        this.manager = manager;
        this.j2oClassManager    = manager.getClassManager();
        this.j2oOntologyManager = manager.getOntologyManager();
        logger = manager.getLogger();}
   
    

      
    // ------------------------------- Object level synchronisation -------------------------- //
    
    
    /** activates synchronisation for the objects in all given individual wrappers. 
     * 
     * @param iwrps are single individual wrappers or arrays or collections of individual wrappers.
     * @param log if 'true' then the activation is logged.
     */
    public void activateSynchronisation(final boolean log, Object... iwrps) {
        for(Object o: iwrps) {
            Utilities.map(o,
                new Function() {
                    @Override
                    public boolean apply(Object object) {
                        ((IndividualWrapper)object).activateSynchronisation(log);
                        return true;}});}}
    
    
       /** deactivates synchronisation for the objects in all given individual wrappers. 
     * 
     * @param iwrps are single individual wrappers or arrays or collections of individual wrappers.
     * @param log if 'true' then the deactivation is logged.
     */
    public void deactivateSynchronisation(final boolean log, Object... iwrps) {
        for(Object o: iwrps) {
            Utilities.map(o,
                new Function() {
                    @Override
                    public boolean apply(Object object) {
                        ((IndividualWrapper)object).deactivateSynchronisation(log);
                        return true;}});}}
    
    
       
   
     // ------------------------------ Synchronisation End ------------------------------------ //
    
    /** clears all internal data about individuals.*/
    public void clear() {
        obj2iwr.clear();
        ind2iwr.clear();}
    
    
    /** returns a set of stored IndividualWrappers for the given objects together with the not-mapped objects.
     * 
     * @param object, the objects are either atomic objects, or arrays, or collections of atomic objects.
     * @return [individualMetadata,unmappedObjects]
     */
    
    Object[] getIWRs(Object... object) {
        final Set<IndividualWrapper> iwrs = new HashSet<IndividualWrapper>();
        final Set<Object> unmapped = new HashSet<Object>();
        for(Object o: object) {
            Utilities.map(o,
                new Function() {
                    @Override
                    public boolean apply(Object object) {
                        IndividualWrapper imd = getIWR(object);
                        if(imd == null) unmapped.add(object);
                        else iwrs.add(imd);
                        return true;}});}
       return new Object[]{iwrs,unmapped};}
    
  
   
    /** turns the object into OWLNamedIndividuals.
     * 
     * @param object a Java object, collection or an array.
     * @return a set of OWLNamedIndividuals.
     * 
     * <p> The object which have already been turned into OWL-individuals are just mapped to their corresponding OWL-individuals.
     * For the other objects, new OWL-individuals are created. 
     * If the mapping to OWL-individuals failed, an error message is logged, but nothing else happens.
     */
    Set<OWLNamedIndividual> makeIndividuals(Object object) {
        final Set<OWLNamedIndividual> individuals = new HashSet<OWLNamedIndividual>();
        Utilities.map(object,
                new Function() {
                    @Override
                    public boolean apply(Object object) {
                        IndividualWrapper iwr = getIWR(object);
                        if(iwr == null) iwr = object2OWL(object);
                        if(iwr != null) individuals.add(iwr.getIndividual());
                        return true;}});
       return individuals;}
   
    /** transfers the objects into individuals and adds them as role fillers for the property.
     * 
     * @param individual the OWL individual for which the role fillers are to be added. 
     * @param object     a single object, an array or collection of objects to be mapped to OWL individuals. 
     * @param property   an OWL object property. 
     */
    private void addObjectPropertyValue(final OWLNamedIndividual individual, final OWLObjectProperty property,Object object) {
        Utilities.map(object,
                new Function() {
                    @Override
                    public boolean apply(Object object) {
                        IndividualWrapper iwr = getIWR(object);
                        if(iwr == null) iwr = object2OWL(object);
                        if(iwr != null) j2oOntologyManager.addObjectPropertyValue(individual,property,iwr.getIndividual());
                        return true;}});}
    
    
    /** transfers the objects into individuals and removes them as role fillers for the property.
     * 
     * @param individual the OWL individual for which the role fillers are to be removed. 
     * @param object     a single object, an array or collection of objects to be mapped to OWL individuals. 
     * @param property   an OWL object property. 
     */
    private void removeObjectPropertyValue(final OWLNamedIndividual individual, final OWLObjectProperty property, Object object) {
        Utilities.map(object,
                new Function() {
                    @Override
                    public boolean apply(Object object) {
                        IndividualWrapper iwr = getIWR(object);
                        if(iwr == null) iwr = object2OWL(object);
                        if(iwr != null) j2oOntologyManager.removeObjectPropertyValue(individual,property,iwr.getIndividual());
                        return true;}});}
    
    
    
    
   
 
    
    /** turns the object into an OWL-individual. 
     * 
     * @param object A Java object which corresponds to an OWL-individual.
     * @return The corresponding OWL-individual.
     * <p>If there is no such individual already, a new one is created.
     */
    private OWLNamedIndividual makeIndividual(Object object) {
        if(object == null) return null;
        IndividualWrapper imd = getIWR(object);
        if(imd == null) imd = object2OWL(object);
        return imd.getIndividual();}

    
    /** local list of OWL axioms to be used in object2OWL */
    private ArrayList<OWLAxiom> axioms = new ArrayList<OWLAxiom>();
    
    /** generates for a Java-object the corresponding OWL-individual.
     * 
     * @param object an atomic Java object (no array or collection)
     * @return the corresponding IndividualWrapper or null if something went wrong.
     * <p>
     * All attributes of the object, which are accessible by the annotated getter methods are also
     * mapped to OWL-individuals.
     * The axioms specifying the new OWL-individuals are generated, but not yet made effective. 
     * Therefore you can generate mappings for several Java-objects first and then call applyChanges.
     * If the generation of the mapping fails at some point, no axioms are inserted into the ontology manager.
     * 
     */
    public IndividualWrapper object2OWL(Object object) {
        if(object == null) return null;
        IndividualWrapper iwr = getIWR(object);
        if(iwr != null) return iwr;   // object is already mapped.
        
        logger.entering("J2OIndividualManager", "object2OWL", object);
        axioms.clear();
        iwr = objects2OWL(object); 
        if(iwr != null) j2oOntologyManager.addOWLAxioms(axioms);
        
        logger.exiting("J2OIndividualManager", "object2OWL", object, iwr);
        return iwr;}
    
    
    /** This is the recursive method which generates for a Java-object the corresponding OWL-individual.
     * 
     * @param object an atomic Java object (no array or collection).
     * @return the corresponding IndividualWrapper or null if something goes wrong. 
     * <p>
     * This method generates the OWL-individual for the object itself and calls 
     * getObjectPropertyAxioms for the relation-type attributes.
     */
    private IndividualWrapper objects2OWL(Object object)   {
        if(object == null) return null;
        IndividualWrapper iwr = getIWR(object);
        if(iwr != null) return iwr;   // object is already mapped.
        
        Class jcl = object.getClass();
        ClassWrapper cmd = j2oClassManager.getCLW(jcl,true,true);
        if(cmd == null) {
            logger.error("Unmapped class", Logging.Origin.USAGE,jcl);
            return null;}   // unknown OWL-class for the object's class.
        
        jcl = cmd.getJavaClass(); // could be a superclass of object which is mapped to an OWL-class.
        OWLClass ocl = cmd.getOWLClass();
        OWLNamedIndividual individual = j2oOntologyManager.createOWLIndividual(object,ocl);
        logger.event(Logging.Semantics.OIC,individual);
        iwr = new IndividualWrapper(new Object[]{object},individual,manager);
        insertIWR(iwr);
        logger.correlation(Logging.Semantics.OIC, object,iwr);
        axioms.add(J2OOntologyManager.getMembershipAxiom(individual,ocl)); // makes individual elementOf ocl
        
        Object value = null;    // now we treat the attributes
        for(PropertyWrapper prwp : cmd.getProperties()) {
            Method getter = prwp.getGetter(jcl);
            try{value = getter.invoke(object);}
            catch(Exception ex) {
                logger.error("Getter not applicable", Logging.Origin.USAGE, getter,object);
                return null;}
            if(prwp.isMapper()) {map2OWL(jcl,individual,prwp,(Map)value); continue;}
            
            if(prwp.isDataProperty()) {
                if(!j2oOntologyManager.makeDataPropertyAxioms(individual,(OWLDataProperty)(prwp.getProperty()),value,axioms)) {
                    logger.error(prwp.getName() +"(" + individual.toString() + "," + value.toString() + ") could not become role-filler(s)",Logging.Origin.USAGE);
                    return null;}}
            else addObjectPropertyAxioms(individual,(OWLObjectProperty)(prwp.getProperty()),value);}
        return iwr;} 

    /** turns the result of a getter method which is a map into OWL axioms */
    private void map2OWL(Class jcl, OWLNamedIndividual individual, PropertyWrapper prwp, Map map) {
        boolean isDataProperty = prwp.isDataProperty();
        for(Object key : map.keySet()) {
            String propertyName = prwp.getName() + "_" + Utilities.replaceWhitespace(key.toString(),'_');
            OWLProperty property = j2oOntologyManager.makeOWLProperty(propertyName,isDataProperty);
            j2oOntologyManager.addOWLAxioms(prwp.createOWLAxiomsForMappers(jcl,property));
            Object value = map.get(key);
            if(isDataProperty) {
                if(!j2oOntologyManager.makeDataPropertyAxioms(individual,(OWLDataProperty)property,value,axioms)) {
                    logger.error(prwp.getName() +"(" + individual.toString() + "," + value.toString() + 
                            ") could not become role-filler(s)",Logging.Origin.USAGE);
                    return;}}
            else addObjectPropertyAxioms(individual,(OWLObjectProperty)property,value);}}
            
    
    
    /** This method adds the object property axioms for the given individual and the property.
     * 
     * @param individual    an OWLIndividual
     * @param property      an OWLObjectProperty
     * @param value         the value of the property. 
     * <p>The value may be an atomic object, an array or a collection. 
     * In the latter cases all elements are added as role-fillers for the OWL-property. 
     * The values are recursively mapped to OWL-individuals. 
     */ 
    private void addObjectPropertyAxioms(final OWLIndividual individual, final OWLObjectProperty property, Object value)  {     
        assert(property != null);
        Utilities.map(value, 
                new Function() {
                    @Override
                    public boolean apply(Object filler) {
                        IndividualWrapper imd = objects2OWL(filler);
                        if(imd == null)  return false;
                        axioms.add(J2OOntologyManager.getObjectPropertyAxiom(individual, property, imd.getIndividual()));
                        return true;}});}
        
        
   
    // ---------------------- BEGIN: Methods called by injected synchroniser code --------------------------//
     
    /** This method forwards in an adder method the information that an object is to be added.
     * 
     * @param object a Java-object whose attribute value is to be inserted/removed
     * @param getter the name of the getter method
     * @param value  the atomic value to be inserted.
     * @return the individual manager. 
     * <p>This method is called by the injected synchroniser code in add-methods for attribute values
     * and updates the extension ontology.
     */
    public J2OIndividualManager addPropertyValue(Object object, Method getter, Object key, Object value) {
        IndividualWrapper iwr = getIWR(object);
        if(iwr == null) {
            logger.error("Unmapped object", Logging.Origin.USAGE,object,this);
            return null;}
        PropertyWrapper pmd = j2oClassManager.getPRW(getter);
        if(pmd == null) {
            logger.error("Unmapped method", Logging.Origin.USAGE, getter);
            return null;} 
        OWLNamedIndividual individual = iwr.getIndividual();
        
        synchronized(iwr) {
            if(pmd.isMapper()) {
                //check whether the call of an adder-method overwrites a key.
                if(value instanceof Collection || value.getClass().isArray()) {
                    Object mapped = null;
                    try{mapped = ((Map)(getter.invoke(object))).get(key);}
                    catch(Exception ex) {
                        logger.error("getter invocation", Logging.Origin.USAGE,object,getter,ex); }
                    if(mapped != null) {removePropertyValue(object,getter,key,mapped);}}
                
                if(pmd.isDataProperty()) {
                    OWLDataProperty dataProperty = (OWLDataProperty)j2oOntologyManager.makeOWLProperty(pmd.getName()+"_" + key.toString(), true);
                    j2oOntologyManager.changeDataProperty(individual, dataProperty, value, true);}
                else {
                    OWLObjectProperty objectProperty = (OWLObjectProperty)j2oOntologyManager.makeOWLProperty(pmd.getName()+"_" + key.toString(),false);
                    addObjectPropertyValue(individual,objectProperty,value);}}
            else {
                value = key;
                if(pmd.isDataProperty()) 
                    j2oOntologyManager.changeDataProperty(individual, (OWLDataProperty)(pmd.getProperty()), value, true);
                else {
                    IndividualWrapper fillermd = getIWR(value);
                    if(fillermd == null) fillermd = object2OWL(value);
                    if(fillermd == null) return null;
                    j2oOntologyManager.addObjectPropertyValue(individual,(OWLObjectProperty)(pmd.getProperty()),fillermd.getIndividual());}}}
        return this;}
    
    
    
    
    /** This method can be called when an already mapped Java-object removes an attribute value.
     * 
     * @param object a Java-object whose attribute value is to be inserted/removed
     * @param getter the name of the getter method
     * @param value  either the atomic value to be removed, or the key if the property is a functional mapper.
     * @return the individual manager. 
     * <p>This method is called by the injected synchroniser code in remove-methods for attribute values
     * and updates the extension ontology.
     * <p>If the property is relational and not a mapper, then the method is called within a remove method: removeItem(item).<br/>
     * If the property is a functional mapper then the method is called within a remove method removeItem(key),
     * where the actual object to be removed is getter().get(key).
     * 
     */
    public J2OIndividualManager removePropertyValue(Object object, Method getter, Object key, Object value) {
        IndividualWrapper iwr = getIWR(object);
        if(iwr == null) {
            logger.error("Unmapped object", Logging.Origin.USAGE, object,this);
            return null;}
        PropertyWrapper pmd = j2oClassManager.getPRW(getter);
        if(pmd == null) {
            logger.error("Unmapped method", Logging.Origin.USAGE, getter);
            return null;} 
        OWLNamedIndividual individual = iwr.getIndividual();
        OWLDataProperty dataProperty = null;
        OWLObjectProperty objectProperty = null;
        
        if(pmd.isMapper()) {
            if(pmd.isFunctional()) {
                Map map = null;
                try {map = (Map)(getter.invoke(object));} 
                catch(Exception ex) {
                    logger.error("Getter Method cannot be invoked.", Logging.Origin.USAGE, getter, object, ex);
                    return null;}
                if(map == null) return this;
                value = map.get(key);
                if(value == null) return this;}
            if(pmd.isDataProperty())  {
                dataProperty = j2oOntologyManager.getFirstOWLDataProperty(pmd.getName()+"_" + key.toString());
                if(dataProperty == null) return this;}
            else {objectProperty = j2oOntologyManager.getFirstOWLObjectProperty(pmd.getName()+"_" + key.toString());
                  if(objectProperty == null) return this;}}
        else {
            value = key;
            if(pmd.isDataProperty())    dataProperty   = (OWLDataProperty)(pmd.getProperty());
            else                        objectProperty = (OWLObjectProperty)(pmd.getProperty());}
               
        synchronized(iwr) {
            if(pmd.isDataProperty()) j2oOntologyManager.changeDataProperty(individual, dataProperty, value, false);
            else removeObjectPropertyValue(individual,objectProperty,value);}
        return this;}
  
    
    
     /** This method can be called when an already mapped Java-object removes all attribute values.
     * 
     * @param object a Java-object whose attribute values are to be removed.
     * @param getter the name of the getter method
     * @param values  the values to be removed.
     * @return the individual manager. 
     * <p>This method is called by the injected synchroniser code in clearer methods for attribute values
     * and updates the extension ontology.
     */
    public J2OIndividualManager clearPropertyValues(Object object, Method getter, Object values) {
        IndividualWrapper iwr = getIWR(object);
        if(iwr == null) {
            logger.error("Unmapped object", Logging.Origin.USAGE, object,this);
            return null;}
        final PropertyWrapper pmd = j2oClassManager.getPRW(getter);
        if(pmd == null) {
            logger.error("Unmapped method", Logging.Origin.USAGE, getter);
            return null;} 
        final OWLNamedIndividual individual = iwr.getIndividual();
        synchronized(iwr) {
            if(pmd.isMapper()) {
                for(Object key : ((Map)values).keySet()) {
                    if(pmd.isDataProperty()) {
                        OWLDataProperty dataProperty = j2oOntologyManager.getFirstOWLDataProperty(pmd.getName()+"_" + key.toString());
                        j2oOntologyManager.changeDataProperty(individual, dataProperty, ((Map)values).get(key), false);}
                    else {
                       OWLObjectProperty objectProperty = j2oOntologyManager.getFirstOWLObjectProperty(pmd.getName()+"_" + key.toString()); 
                       removeObjectPropertyValue(individual,objectProperty,((Map)values).get(key));}}}
            else {
                if(pmd.isDataProperty()) 
                    j2oOntologyManager.changeDataProperty(individual, (OWLDataProperty)(pmd.getProperty()), values, false);
                else {removeObjectPropertyValue(individual,(OWLObjectProperty)(pmd.getProperty()),values);}}}
        return this;}
    
    /** This method can be called when an already mapped Java-object replaces an attribute value by a new one
     * 
     * @param object a Java-object whose attribute value is to be inserted/removed
     * @param getter a getter-method in the Java-class which is mapped to an OWL-property.
     * @param oldValue  the value to be removed.
     * @param newValue  the value to be inserted.
     * @return the individual manager.
     * <p>This method is called by the injected synchroniser code in setter methods for attribute values. 
     * The given method, however, should be the annotated getter-method. 
     */
    public J2OIndividualManager exchangePropertyValue(Object object, Method getter, Object oldValue, Object newValue)  {
        IndividualWrapper iwr = getIWR(object);
        if(iwr == null) {
            logger.error("Unmapped object", Logging.Origin.USAGE, object,this);
            return this;}
        PropertyWrapper pmd = j2oClassManager.getPRW(getter);
        if(pmd == null) {
            logger.error("Unmapped method", Logging.Origin.USAGE, getter);
            return this;}
        OWLNamedIndividual individual = iwr.getIndividual();
        synchronized(iwr) {
            if(pmd.isMapper()) exchangePropertyMap(individual,pmd, (Map)oldValue, (Map)newValue);
            else {
                if(pmd.isDataProperty()) {
                    j2oOntologyManager.changeDataProperty(individual, (OWLDataProperty)(pmd.getProperty()), oldValue, newValue);
                    return this;}
                else {
                    OWLObjectProperty property = (OWLObjectProperty)(pmd.getProperty());
                    removeObjectPropertyValue(individual,property,oldValue);
                    addObjectPropertyValue(individual,property,newValue);}}}
        return this;}
    
   
    
        /** generates the OWL axioms for exchanging the values of functional maps like Map<Key,Value> or Map<Key,Values>
         * 
         * @param individual the OWL individual for which the values are to be exchanged.
         * @param pmd        the PropertyWrapper.
         * @param oldValue   the old map.
         * @param newValue   the new map.
         */
        private void exchangePropertyMap(OWLNamedIndividual individual, PropertyWrapper pmd, Map oldValue, Map newValue) {
            String name = pmd.getName()+"_";
            if(pmd.isDataProperty()) {
                if(oldValue != null) {
                    for(Object key : oldValue.keySet()) {
                        OWLDataProperty dataProperty = j2oOntologyManager.getFirstOWLDataProperty(name + key.toString());
                        if(dataProperty == null) continue;
                        j2oOntologyManager.changeDataProperty(individual, dataProperty, oldValue.get(key), false);}}
                if(newValue != null) {
                    for(Object key : newValue.keySet()) {
                        OWLDataProperty dataProperty = (OWLDataProperty)j2oOntologyManager.makeOWLProperty(name + key.toString(),true);
                        j2oOntologyManager.changeDataProperty(individual, dataProperty, newValue.get(key), true);}}}
            else {
                if(oldValue != null) {
                    for(Object key : oldValue.keySet()) {
                        OWLObjectProperty objectProperty = j2oOntologyManager.getFirstOWLObjectProperty(name + key.toString());
                        if(objectProperty == null) continue;
                        removeObjectPropertyValue(individual,objectProperty,oldValue.get(key));}}
                if(newValue != null) {
                    for(Object key : newValue.keySet()) {
                        OWLObjectProperty objectProperty = (OWLObjectProperty)j2oOntologyManager.makeOWLProperty(name + key.toString(),false);
                        addObjectPropertyValue(individual,objectProperty,newValue.get(key));}}}}
    
      
     
    // ---------------------- END: Methods called by injected synchroniser code --------------------------//
     
    // ---------------------- BEGIN: Asynchronous Synchronisation --------------------------//
    
    /** synchronises the objects in the given individual wrappers with the corresponding OWL-individual. 
      * <p> This method can be called if the object synchronisation has been turned off for a while.
     * @param iwrps are single individual wrappers, arrays or collection of individual wrappers.
     * @throws ClassClastException if the objects are not individual wrappers.
      * 
      */
    public void synchroniseWithOWL(Object... iwrps) {
        for(Object iwrp : iwrps) {
             Utilities.map(iwrp, 
                new Function() {
                    @Override
                    public boolean apply(Object iwr) {
                        ((IndividualWrapper)(iwr)).synchroniseWithOWL();
                        return true;}});}}
     
     
  
     // ---------------------- END: Asynchronous Synchronisation --------------------------//
     
     
     
    /** sets the value as role-filler for the individual's property.
      * 
      * @param individual an OWLindividual
      * @param property   a OWLObjectPropert
      * @param value  the properties role filler
      * @return true if the operation succeeded.
      * <p> The value is made the individual's role filler. 
      */
    boolean setObjectPropertyValue(OWLNamedIndividual individual, OWLObjectProperty property, Object value) {
        IndividualWrapper filler = object2OWL(value);
        if(filler != null) {
            j2oOntologyManager.setObjectProperty(individual,property,filler.getIndividual());
            return true;}
        else {
            logger.error("The value cannot be mapped to OWL", Logging.Origin.USAGE, individual, property,value);
        return false;}
    }
    
     
    /** gets all the individual's role fillers for the given property. 
     * 
     * @param individual an individual
     * @param property an OWLObjectProperty
     * @return th4e role fillers as array of IndividualProxys.
     */ 
    IndividualWrapper[] getObjectPropertyValues(OWLNamedIndividual individual, OWLObjectProperty property) {
         Set<OWLNamedIndividual> individuals = j2oOntologyManager.getObjectPropertyValues(individual,property);
         if(individuals == null) return new IndividualWrapper[0];  // a reasoner error occurred.
         IndividualWrapper[] indWrappers = new IndividualWrapper[individuals.size()];
         int i = 0;
         for(OWLNamedIndividual ind : individuals)  indWrappers[i] = individual2Java(ind);
         return indWrappers;}
     
     
     
    /** maps an OWL-individual to corresponding Java-objects. 
      * 
      * @param individual an OWLNamedIndividual
      * @return the corresponding IndividualWrapper or null, if, for example the OWL-classes are not mapped to Java classes.
      * Since OWL allows for multiple inheritance, and Java does not, the system may map a single
      * OWL-individual to several Java-objects. 
      * As an example an instance of a amphibian-vehicle (a subclass of Ship and Surface-Vehicle) would
      * be mapped to both, a ship and a surface-vehicle.
      * The correspondences are kept in the IndividualWrapper structure.
      */
    public IndividualWrapper individual2Java(OWLNamedIndividual individual) {
        IndividualWrapper indWrapper = getIWR(individual);
        if(indWrapper != null) return indWrapper;
        try {indWrapper = new IndividualWrapper(individual,manager);}
        catch(J2OException ex) {
            logger.error("OWL->Java mapping error", Logging.Origin.USAGE, ex);   
            return null;}
        insertIWR(indWrapper);
        return indWrapper;}
    
   
    /** reclassifies all objects in the given individual wrappers.
     * 
     * @param iwrds single individual wrappers, arrays or collections of individual wrappers.
     * @param forceReclassification if this flag is false then reclassification is done only if some 
     * attributes have been changed since the last reclassification.
     * @throws ClassCastException if the objects are not IndividualWrappers.
     */
    public void reclassify(final boolean forceReclassification, Object... iwrds) {
        for(Object iwrd: iwrds) {
            Utilities.map(iwrd,
                new Function() {
                    @Override
                    public boolean apply(Object object) {
                        ((IndividualWrapper)(object)).reclassify(forceReclassification); 
                        return true;}});}}
        
    /** this is used to ensure pointer-equal attributes of the same proprties during reclassification. */
    private HashMap<PropertyWrapper,Object> propertyValueCache = new HashMap<PropertyWrapper,Object>();
    
    /** clears the cache. */
    void clearPropertyValueCache() {propertyValueCache.clear();}
    
    /** transfers the properties of an OWL individual to a Java object.
     * 
     * @param object a newly generated Java object
     * @param jcl    the object's class or superclass
     * @param individual the OWL individual
     * @param  referenceObject is another object which may share fields with the new object.
     * <p>This method follows the Java class hierarchy upwards and starts transferring
     * properties from the topmost annotated class.
     * <p> Common fields between the object and reference object are just transferred from the reference 
     * object to the new object.
     */
    
    boolean transferOWLProperties(OWLNamedIndividual individual, Object object, Class jcl, boolean forceReclassification) {
        boolean synchronised = false;
        try{synchronised = Utilities.getAndSetObjectFlag(object, J2OSynchroniser.SYNCHRONIZE_BYOBJECT_VARIABLE, IndividualWrapper.synchronisationFlag);}
        catch(Exception ex){}
        
        Class superclass = jcl.getSuperclass();
        if(superclass != null) transferOWLProperties(individual,object,superclass,forceReclassification);
        ClassWrapper clwrp = j2oClassManager.getCLW(jcl);
        if(clwrp == null) return true;
        boolean result = transferLocalOWLProperties(individual,object,clwrp,forceReclassification);
        
        if(!synchronised) {
            try{Utilities.resetObjectFlag(object, J2OSynchroniser.SYNCHRONIZE_BYOBJECT_VARIABLE, IndividualWrapper.synchronisationFlag);}
            catch(Exception ex){}}
        return result;}
    
    
    /** transfers those properties of an OWL individual to a Java object which are local to the Java object.
     *<p>There is a nondeterminism in this algorithm:<br/>
     * Suppose the individual I has an object property P with filler F (an OWL individual)
     * Suppose F is mapped to two or more Java Objects o1,o2,...
     * If property P has a setter S with argument type T, then it is not clear which of the
     * objects o1,o2,... to be filled into S. <br/>
     * The algorithm chooses the first one whose type is T or a subtype of T. <br/>
     * If none of the objects has an appropriate type then an error is generated.
     * 
     * <p>If forceReclassification = true then there may be  changes in the OWL ontology  
     * which are not directly related with the individual, but the reasoner can infer information relevant for the individual. 
     * In this case all attributes must be updated. 
     */
    private boolean transferLocalOWLProperties(OWLNamedIndividual individual, Object object, ClassWrapper clwrp, boolean forceReclassification) {
        Class jcl = clwrp.getJavaClass();
        for(PropertyWrapper pmd : clwrp.getProperties()) {
            OWLProperty property = pmd.getProperty();
            PropertyWrapperComponent pwc = pmd.getPropertyWrapperComponent(jcl);
            Method setter = pwc.setter;
            Method adder = pwc.adder;
            
            // even if forceReclassification = true we want to ensure pointer equal values for 
            // the same property of different objects in the same individual wrapper. 
            if(forceReclassification && setter != null) {
                Object cached = propertyValueCache.get(pmd);
                if(cached != null) {
                    try {setter.invoke(object,cached);}
                    catch (Exception ex) {
                        logger.error("Setter failure", Logging.Origin.USAGE, setter, jcl, object, cached, pmd,ex);
                        return false;}
                    continue;}}
            
            Method getter =  pwc.getter;
            boolean functional = pmd.isFunctional();
            Object contents = null;
            
            
            if(!forceReclassification) {// if the attribute is already filled we do nothing 
                try{contents = getter.invoke(object);}
                catch(Exception ex) {
                    logger.error("Getter cannot be invoked", Logging.Origin.USAGE, object,jcl,getter,ex);
                    continue;}
                if(!Utilities.isNull(contents)) continue;}
             
            // already filled attributes must be cleared.
            if(forceReclassification && !functional) {
                Method clearer = pwc.clearer;
                if(clearer == null) {
                    logger.error("No clearer defined",Logging.Origin.JAVA,getter);
                    return false;}
                else 
                    try{clearer.invoke(object);}
                    catch (Exception ex) {
                        logger.error("Clearer failure", Logging.Origin.USAGE, clearer, jcl, object, pmd,ex);
                        return false;}}
           
            if(!functional) setter = pwc.adder;
            if(setter == null) continue;
            
            if(pmd.isMapper()) {
                Method keyConstructor = pwc.getKey;
                if(pmd.isDataProperty()) {
                    for(OWLDataProperty dataProperty: j2oOntologyManager.findOWLDataProperties(pmd.getName()+"_")) {
                        String fragment = dataProperty.getIRI().getFragment();
                        String keyString = fragment.substring(fragment.indexOf("_")+1,fragment.length());
                        Object keyObject = keyString;
                        if(keyConstructor != null) {
                            try {keyObject = keyConstructor.invoke(object,keyString);}
                            catch(Exception ex) {
                                logger.error("Key Constructor failed", Logging.Origin.USAGE, keyConstructor, jcl, object, pmd,ex);
                                return false;}}
                        Set<Object> values = j2oOntologyManager.getDataPropertyValues(individual,dataProperty);
                        if(values == null) return false; // the reasoner caused an error.
                        if(functional && values.size() > 1) {
                            logger.error("Functional Property has more than one argument",Logging.Origin.USAGE, individual,dataProperty);
                                return false;}
                        for(Object value : values) {
                            try {adder.invoke(object, keyObject,value);} 
                            catch (Exception ex) {
                                logger.error("Adder failure", Logging.Origin.USAGE, adder, jcl, object, value, pmd,ex);
                                return false;}}}}
                    
                else {
                    for(OWLObjectProperty objectProperty: j2oOntologyManager.findOWLObjectProperties(pmd.getName()+"_")) {
                        String fragment = objectProperty.getIRI().getFragment();
                        String keyString = fragment.substring(fragment.indexOf("_")+1,fragment.length());
                        Object keyObject = keyString;
                        if(keyConstructor != null) {
                            try {keyObject = keyConstructor.invoke(keyString);}
                            catch(Exception ex) {
                                logger.error("Key Constructor failed", Logging.Origin.USAGE, keyConstructor, jcl, object, pmd,ex);
                                return false;}}
                        Set<OWLNamedIndividual> values = j2oOntologyManager.getObjectPropertyValues(individual,objectProperty);
                        if(values == null) return false; // the reasoner caused an error.
                        if(functional && values.size() > 1) {
                            logger.error("Functional Property has more than one argument", Logging.Origin.USAGE, individual,property);
                            return false;}
                        
                        for(OWLNamedIndividual value : values) { 
                            IndividualWrapper vimd = individual2Java(value);
                            if(vimd == null) continue;
                            boolean found = false;
                            for(Object o : vimd.getObjects()) { 
                                try {setter.invoke(object, keyObject, o);}
                                catch (Exception ex) {continue;}
                                found = true; // here is some nondeterminism. 
                                break;}       // it is not really clear which object to choose.
                            if(!found) {
                                logger.error("Setter failure", Logging.Origin.USAGE, setter, jcl, object, pmd);
                                return false;}}}}}
                
            else {
                if(pmd.isDataProperty()) {
                    Set<Object> values = j2oOntologyManager.getDataPropertyValues(individual,(OWLDataProperty)property);
                    if(values == null) return false; // the reasoner caused an error.
                    if(values.isEmpty()) continue;
                    if(functional && values.size() > 1) {
                        logger.error("Functional Property has more than one argument",Logging.Origin.USAGE, individual,property);
                        return false;}
                    for(Object value : values) {
                        try {setter.invoke(object, value);} 
                        catch (Exception ex) {
                            logger.error("Setter failure", Logging.Origin.USAGE, setter, jcl, object, value, pmd,ex);
                            return false;}}}
                else {
                    Set<OWLNamedIndividual> values = j2oOntologyManager.getObjectPropertyValues(individual,(OWLObjectProperty)property);
                    if(values == null) return false; // the reasoner caused an error.
                    if(values.isEmpty()) continue;
                    if(functional && values.size() > 1) {
                        logger.error("Functional Property has more than one argument", Logging.Origin.USAGE, individual,property);
                        return false;}
                    for(OWLNamedIndividual value : values) { 
                        IndividualWrapper vimd = individual2Java(value);
                        if(vimd == null) continue;
                        boolean found = false;
                        for(Object o : vimd.getObjects()) { 
                            try {setter.invoke(object, o);}
                            catch (Exception ex) {continue;}
                            found = true; // here is some nondeterminism. 
                            break;}       // it is not really clear which object to choose.
                        if(!found) {
                            logger.error("Setter failure", Logging.Origin.USAGE, setter, jcl, object, pmd);
                            return false;}}}}
            
            if(forceReclassification) {
                try{contents = getter.invoke(object);}
                catch(Exception ex) {logger.error("Getter cannot be invoked", Logging.Origin.USAGE, object,jcl,getter,ex);}
                if(contents != null) propertyValueCache.put(pmd,contents);}
        }
        return true;}
      
    
    /** uses the reasoner to find all instances of the class expression and maps them back to the Java-objects.
     * 
     * @param cle an OWLClassExpression.
     * @return a list of IndividualWrapper which represent the Java-objects corresponding to the instances of class expression.
     */
    public ArrayList<IndividualWrapper> getInstances(OWLClassExpression cle) {
        Set<OWLNamedIndividual> individuals = j2oOntologyManager.getInstances(cle);
        ArrayList<IndividualWrapper> objects = new ArrayList<IndividualWrapper>(individuals.size());
        for(OWLNamedIndividual individual : individuals) {
            IndividualWrapper imd = getIWR(individual);
            if(imd == null) {logger.error("unmapped individual",Logging.Origin.USAGE, individual);}
            else objects.add(imd);}
        return objects;}
    
    /** uses the reasoner to find all instances of the class expression and maps them back to the Java-objects.
     * 
     * @param cle an OWLClassExpression in Manchester Syntax
     * @return a list of IndividualWrapper which represent the Java-objects corresponding to the instances of class expressions.
     */
     public ArrayList<IndividualWrapper> getInstances(String cle) throws ParserException {
         return getInstances(j2oOntologyManager.parseClassExpression(cle));
     }
    
    
     
     /** @return a String-representation of the entire data. */
    @Override
     public String toString() {
         String blanks = "    ";
         String s = "J2OIndividualManager " + getName() +  "\n";
         s += "\n\nObjects -> Individual:\n";
         for(Entry<Object,IndividualWrapper> entry : obj2iwr.entrySet()) 
             s += blanks + entry.getKey().toString() + " -> " + entry.getValue().getIndividual().toString() + "\n";
         s += "\nIndividual -> Objects:\n";
         for(Entry<OWLNamedIndividual,IndividualWrapper> entry : ind2iwr.entrySet())
             s += blanks + entry.getKey().toString() + " -> " + entry.getValue().getObjectsString() + "\n";        
         return s;
     }
    
    public static void main(String[] args) {
        
    }
}