 /*
    * Copyright (c) 2012, H. J. Ohlbach. All rights reserved.
    * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
    *
    * This code is free software; you can redistribute it and/or modify it
    * under the terms of the GNU General Public License version 2 only, as
    * published by the Free Software Foundation. 
    *
    * This code is distributed in the hope that it will be useful, but WITHOUT
    * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    * version 2 for more details.
    */

package Java2OWL;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.lang.instrument.IllegalClassFormatException;
import java.security.ProtectionDomain;
import java.lang.instrument.ClassFileTransformer;
import java.util.ArrayList;
import java.util.Map;
import org.apache.bcel.*;
import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;

/** This is the program which manipulates the annotated compiled class files to 
 * insert the code for synchronising Java and OWL. 
 * <p> An example is: The class <br/>
 * <pre>
 * &#064;J2OWLClass(name = "Person", OWLSuperClasses = "LivingThing")
 * class Person {
 * boolean male; 
 * public TestPerson(boolean male) {this.male = male;}
 * 
 * &#064;J2OWLProperty(setter = "setGender", required=true)
 * public boolean isMale(){return male;}
 * 
 * public void setGender(boolean male){
 *      this.male = male;}}
 * </pre>
 * 
 * would be changed to 
 * <pre>
 * &#064;J2OWLClass(name = "Person", OWLSuperClasses = "LivingThing")
 * class Person {
 * boolean male;  
 * public static J2OIndividualManager j2OWLIndividualManager;
 * public boolean j2OWLNoSynchronizationByObject;
 * private static Method J2OWL_getter_isMale = J2OSynchroniser.getMethod(TestPackage/TestPerson, "isMale");
 * 
 * J2OIndividualManager J2OSynchronisationActive() {
 *    Java2OWLFlags |= 6;
 *    return  (Java2OWLFlags & 1) == 0 ? null : j2OWLIndividualManager;}
 * 
 * public TestPerson(boolean male) {this.male = male;}
 * 
 * &#064;J2OWLProperty(setter = "setGender", required=true)
 * public boolean isMale(){return male;}
 * 
 * public void setGender(boolean male){
 * if(J2OSynchronisationActive() != null)
 *      J2OSynchronisationActive().exchangePropertyValue(this, J2OWL_getter_isMale, Boolean.valueOf(isMale()), Boolean.valueOf(male));
 *      this.male = male;}}
 * </pre>
 * <p> 
 * The variables j2OWLIndividualManager and Java2OWLFlags
 * are injected in each class even in subclasses. 
 * This means, if A is a subclass of B then A has these two variables twice,
 * its own pair and the inherited pair.
 * The method J2OSynchronisationActive() accesses the pair belonging to the object's actual class.
 * Therefore one can, for example, activate synchronisation for a superclass and deactivate it for a subclass.
 *
 * @author ohlbach
 */
public class J2OSynchroniser implements ClassFileTransformer {
    
    private static String classAnnotation =  "LJava2OWL/J2OWLClass;";
    private static String methodAnnotation =  "LJava2OWL/J2OWLProperty;";
    
    /** For test purposes: a filename where the transformed class is to be dumped. */
    public String dump = null; 
    
    /** creates a new J2OSynchroniser */
    public J2OSynchroniser() {
        super();
        Attribute.addAttributeReader("RuntimeVisibleAnnotations",new AnnotationReader());
    }

    /** This method is called by the class loader. It initiates the transformation */
    @Override
    public byte[] transform(ClassLoader loader, String className, Class redefiningClass, ProtectionDomain domain, byte[] bytes) throws IllegalClassFormatException {
        return injectSynchronizer(className,bytes);}
    
    static final String INDIVIDUAL_MANAGER_VARIABLE = "j2OWLIndividualManager";
    static final String SYNCHRONIZE_BYOBJECT_VARIABLE  = "Java2OWLFlags";
    private static final String[] SYNCHRONIZER_METHODS = new String[]{"exchangePropertyValue","addPropertyValue","removePropertyValue","clearPropertyValues"};
    private static final Type[][] SYNCHRONIZER_METHOD_TYPES = 
            new Type[][]{new Type[]{Type.OBJECT, Type.getType(java.lang.reflect.Method.class), Type.OBJECT, Type.OBJECT}, // setter
                         new Type[]{Type.OBJECT, Type.getType(java.lang.reflect.Method.class), Type.OBJECT, Type.OBJECT}, // adder
                         new Type[]{Type.OBJECT, Type.getType(java.lang.reflect.Method.class), Type.OBJECT, Type.OBJECT}, // remover
                         new Type[]{Type.OBJECT, Type.getType(java.lang.reflect.Method.class), Type.OBJECT}};             // clearer
            
   
    
    /** generates a name for the private static getter-variables which are bound to the getter-methods 
     These names should not be used anywhere else in the class body.*/ 
    private static String getterVariable(Method getter) {
        return "J2OWL_getter_" + getter.getName();}
    
    /** This method initiates all the transformations at byte-code level.
     * 
     * @param className the name of the class 
     * @param bytes the contents of the class-file 
     * @return the transformed contents of the class-file or null if no transformation is necessary.
     * <p>
     * Class files without the Java2OWL/J2OWLClass-annotation are not transformed<br/>
     * Class files with this annotation, but with attribute "synchronize=false" are not transformed.
     */
    private byte[] injectSynchronizer(String className, byte[] bytes) {
        JavaClass jclb = null;
        try{jclb = new ClassParser(new ByteArrayInputStream(bytes),"").parse();}
        catch(IOException ex) {ex.printStackTrace(); return null;}
        for(Attribute attribute : jclb.getAttributes()) {
            if(attribute instanceof AnnotationsAttribute) {
                Map annotation = ((AnnotationsAttribute)attribute).getAnnotation(classAnnotation);
                if(annotation == null) return null;
                Integer synch = (Integer)annotation.get("synchronise");
                if(synch != null && (int)synch == 0) return null;
                return injectSynchronizer(jclb,className);}}
        return null;}
    
    
    /** injects the synchroniser code into a .class file.
     * 
     * @param className The name of the class to inject the synchroniser code into. 
     * @throws ClassNotFoundException
     * @throws FileNotFoundException
     * @throws ClassNotFoundException, FileNotFoundException, IOException 
     * <p>
     * The transformed class overwrites the original .class file.<br/>
     * Class files without the Java2OWL/J2OWLClass-annotation are not transformed<br/>
     * Class files with this annotation, but with attribute "synchronize=false" are not transformed.
     */
    
    void injectSynchronizer(String className) throws ClassNotFoundException, FileNotFoundException, IOException {
        JavaClass jclb = Repository.lookupClass(className);
        for(Attribute attribute : jclb.getAttributes()) { 
            if(attribute instanceof AnnotationsAttribute) {
                Map annotation = ((AnnotationsAttribute)attribute).getAnnotation(classAnnotation);
                if(annotation == null) return;
                Integer synch = (Integer)annotation.get("synchronise");
                if(synch != null && (int)synch == 0) return;
                byte[] bytes = injectSynchronizer(jclb,className);
                if(bytes == null) return;
                File file = new File(Repository.lookupClassFile(className).getPath());
                OutputStream out = new FileOutputStream(file);
                out.write(bytes);
                out.close();}}} 
   
    /** tests whether a Java class is synchronised */
    static boolean isSynchronised(Class jcl){
        try {jcl.getDeclaredField(INDIVIDUAL_MANAGER_VARIABLE);} 
        catch (NoSuchFieldException ex) {return false;}
        catch (SecurityException ex) {return false;}
        return true;}
    
    
     /** This method actually injects the synchroniser code into the class files. 
      */
    private byte[] injectSynchronizer(JavaClass jclb, String className)  {
        
        for(Field  field : jclb.getFields()) {  //check if the class has already been transformed.
            if(field.getName().equals(INDIVIDUAL_MANAGER_VARIABLE)) return null;}
        System.out.println("Transform Class: " + className); 
        ClassGen  jclg = new ClassGen(jclb);
        ConstantPoolGen cp = jclg.getConstantPool();
        InstructionFactory factory = new InstructionFactory(jclg, cp);
        ArrayList<Method> getters = findGetters(jclg); 
        
        // Now we inject the two  fields 
        // public static J2OIndividualManager j2OWLIndividualManager;
        // public boolean j2OWLNoSynchronizationByObject;
        // and all the getter-variables. 
        injectFields(jclg,className,cp,getters);
        createSyncTest(className, jclg, factory, cp);
        
        // Now we must create/extend the <clinit> method to add sttic code for initializing the getter-variables. 
        Method clinit = findCLINIT(jclg);
        if(clinit == null) createCLINIT(jclg, className, getters, factory, cp);
        else {
            MethodGen mg = new MethodGen(clinit,className,cp);
            extendCLINIT(className, mg, getters, factory, cp);
            jclg.replaceMethod(clinit, mg.getMethod());}
        
        // Now we inject the synchronizer code.
        // Example: if(j2oindividualmanager == null)
        //          j2OWLIndividualManager.removePropertyValue(this, J2OWL_getter_hasFriends, friend);

        for(Method method : jclg.getMethods()) {
            Object[] result = isChanger(method,getters);
            if(result == null) continue;
            Method getter = (Method)(result[0]);
            int version = (Integer)(result[1]);
            MethodGen mg = new MethodGen(method,className,cp);
           
            extendInstructionList(jclg,className,mg,getter,version,cp,factory);
            jclg.replaceMethod(method, mg.getMethod());}
        
        // Now we turn the manipulated code back into an ordinary Java-class
        jclb = jclg.getJavaClass(); 
        
        // This is for test purposes.
        if(dump != null) {
            try {
                FileOutputStream st = new FileOutputStream(dump);
                jclb.dump(st);
                System.out.println("Dumped to " + dump);
            }
            catch(Exception ex) {ex.printStackTrace();}}
        // return the class files as a sequence of bytes
        return jclb.getBytes();
    }
         
    
    /** checks whether the method is a setter, adder, remover of clearer
     * 
     * @param getters a list of getter-methods (needed for their annotations)
     * @param method a BCEL-method
     * @return [getter,version] or null of it is no such method.
     * <p>The version is:
     * <ol>
     * <li> setter </li>
     * <li> adder </li>
     * <li> remover </li> 
     * <li> clearer </li>
     * </ol>
     */
    
    private Object[] isChanger(Method method, ArrayList<Method> getters) {
        String name = method.getName();
        String regexpr = name + ".*|.*,\\s*"+name+"\\s*,.*|.*"+name;
        //The annotations may be comma separated lists of setter, adder, remover names.
        for(Method getter : getters) {
            for(Attribute attribute : getter.getAttributes()) {
                if(attribute instanceof AnnotationsAttribute) {
                    Map map = ((AnnotationsAttribute)attribute).getAnnotation(methodAnnotation);
                    String accessor = (String)map.get("setter");
                    if(accessor != null && accessor.matches(regexpr)) {return new Object[]{getter,0};}
                    accessor = (String)map.get("adder");
                    if(accessor != null && accessor.matches(regexpr)) {return new Object[]{getter,1};}
                    accessor = (String)map.get("remover");
                    if(accessor != null && accessor.matches(regexpr)) {return new Object[]{getter,2};}
                    accessor = (String)map.get("clearer");
                    if(accessor != null && accessor.matches(regexpr)) {return new Object[]{getter,3};}}}}
        return null;}
    
    /** finds all annotated getter-methods */
    private ArrayList<Method> findGetters(ClassGen jclg) {
        ArrayList<Method> getter = new ArrayList<Method>();
        for(Method method : jclg.getMethods()) {
            for(Attribute attribute : method.getAttributes()) {
                if(attribute instanceof AnnotationsAttribute &&
                    ((AnnotationsAttribute)attribute).getAnnotation(methodAnnotation) != null){
                    getter.add(method); 
                    break;}}}
        return getter;}
    
    /** searches the <clinit>  method */
    private Method findCLINIT(ClassGen jclg) {
        for(Method method : jclg.getMethods()) {if(method.getName().equals("<clinit>")) return method;}
        return null;}
    
    
    
    /** injects into the java code the fields necessary for synchronising Java with OWL */
    private void injectFields(ClassGen jclg, String className, ConstantPoolGen cp, ArrayList<Method> getters) {
        Type imgType = new ObjectType("Java2OWL.J2OIndividualManager");
        FieldGen field = new FieldGen(Constants.ACC_PUBLIC|Constants.ACC_STATIC, imgType, INDIVIDUAL_MANAGER_VARIABLE, cp);
        jclg.addField(field.getField());
        field = new FieldGen(Constants.ACC_PUBLIC, Type.INT, SYNCHRONIZE_BYOBJECT_VARIABLE, cp);
        jclg.addField(field.getField());
        
        // Now we add static fields like 'private static Method getter_getName' 
        // They are initialized in the <clinit> method with the pointer to the corresponding getter-method.
        // They are needed in the synchronizer calls 
        for(Method getter : getters) {
            field = new FieldGen(Constants.ACC_PRIVATE | Constants.ACC_STATIC, new ObjectType("java.lang.reflect.Method"), getterVariable(getter), cp);
            jclg.addField(field.getField());}}
    
    /** injects the synchroniser code into the instruction list. 
     * <p> the injected code looks roughly like this:
     * <pre>
        J2OIndividualManager j2oindividualmanager = J2OSynchronisationActive();
        j2oindividualmanager;
        if(j2oindividualmanager == null) goto _L2; else goto _L1
    _L1:
        this;
        J2OWL_getter_getName;
        getName();
        name;
        exchangePropertyValue();
    _L2:
        JVM INSTR pop ;
    </pre>
     */
    private void extendInstructionList(ClassGen jclg, String className, MethodGen method, Method getter, int version, 
            ConstantPoolGen cp, InstructionFactory factory) {
        InstructionList il = new InstructionList();
        String getterName = getter.getName();
        int arguments = method.getArgumentTypes().length;
        Type imgType = new ObjectType("Java2OWL.J2OIndividualManager");
        
        InstructionHandle ih_0 = il.append(factory.createLoad(Type.OBJECT, 0));
        il.append(factory.createInvoke(className, "J2OSynchronisationActive", imgType, Type.NO_ARGS, Constants.INVOKEVIRTUAL));
        il.append(InstructionConstants.DUP);
        BranchInstruction ifnull = factory.createBranchInstruction(Constants.IFNULL, null);
        il.append(ifnull);
        il.append(factory.createLoad(Type.OBJECT, 0));
        il.append(factory.createFieldAccess(className, getterVariable(getter), new ObjectType("java.lang.reflect.Method"), Constants.GETSTATIC));
        if(version == 0 || version == 3) injectGetter(il,factory,method,className,getter); // setter, clearer
        if(version != 3) injectArgumentAccess(il,factory,method,0);  //setter,adder,remover
        if(version == 1 || version == 2) {//adder,remover
            if(arguments > 1) {injectArgumentAccess(il,factory,method,1);}
            else il.append(InstructionConstants.ACONST_NULL);}
        il.append(factory.createInvoke("Java2OWL.J2OIndividualManager", SYNCHRONIZER_METHODS[version], imgType, 
                SYNCHRONIZER_METHOD_TYPES[version], Constants.INVOKEVIRTUAL));
   
        InstructionHandle ipop = il.append(InstructionConstants.POP);
        ifnull.setTarget(ipop);
      
        InstructionList ilold = method.getInstructionList();
        ilold.insert(il);
        method.setMaxStack(); 
        method.setMaxLocals(); 
        il.dispose();}
    
    /** injects an access to the getter method in the call to 3rd argument of 
     * j2OWLIndividualManager.exchangePropertyValue(this, J2OWL_getter_isMale, Boolean.valueOf(isMale()), Boolean.valueOf(male));
     */
     private void injectGetter(InstructionList il, InstructionFactory factory, MethodGen method, String className, Method getter) {
        String getterName = getter.getName();
        Type type = getter.getReturnType();
        il.append(InstructionFactory.createLoad(Type.OBJECT, 0));
        if(type instanceof BasicType) { // we need to wrap the value into a wrapper class.
            il.append(factory.createInvoke(className, getterName, type, Type.NO_ARGS, Constants.INVOKEVIRTUAL));
            String wrapper = Utilities.getWrapperType((BasicType)type);
            il.append(factory.createInvoke(wrapper, "valueOf", new ObjectType(wrapper), new Type[] {type}, Constants.INVOKESTATIC));}
        else {
            il.append(factory.createInvoke(className, getterName, type, Type.NO_ARGS, Constants.INVOKEVIRTUAL));}
     }
    /** inserts the last parameter access for a synchroniser call
     * <p> If the call is exchangePropertyValue(this, J2OWL_getter_isMale, isMale(), male);
     * then the access to 'male' is inserted. 
     * <p> If the type is primitive then a conversion to a wrapper-type must be inserted. 
     * Otherwise just a load-command is sufficient. 
      */
    private void injectArgumentAccess(InstructionList il, InstructionFactory factory, MethodGen method, int argument) {
        Type type = method.getArgumentTypes()[argument];
        if(type instanceof BasicType) {
            il.append(InstructionFactory.createLoad(type, argument+1));
            String wrapper = Utilities.getWrapperType((BasicType)type);
            il.append(factory.createInvoke(wrapper, "valueOf", new ObjectType(wrapper), new Type[] {type}, Constants.INVOKESTATIC));}
        else il.append(InstructionFactory.createLoad(type, argument+1));}
    
    
    /** creates a new <clinit> method which binds the static getter-variables to the corresponding getter-methods. 
     * Example:  private static Method getter_getAddress = J2OSynchroniser.getMethod(TestPerson1.class,"getAddress");*/
    private void createCLINIT(ClassGen jclg, String className, ArrayList<Method> getters, InstructionFactory factory, ConstantPoolGen cp) {
        InstructionList il = new InstructionList();
        MethodGen method = new MethodGen(Constants.ACC_STATIC, Type.VOID, Type.NO_ARGS, new String[] {  }, "<clinit>", className, il, cp);
        LDC_W ldc = new LDC_W(cp.lookupClass(className));
        for(Method getter : getters) {
            String name = getter.getName();
            il.append(ldc);
            il.append(new PUSH(cp, name));
            il.append(factory.createInvoke("Java2OWL.J2OSynchroniser", "getMethod", new ObjectType("java.lang.reflect.Method"), new Type[] { new ObjectType("java.lang.Class"), Type.STRING }, Constants.INVOKESTATIC));
            il.append(factory.createFieldAccess(className, getterVariable(getter), new ObjectType("java.lang.reflect.Method"), Constants.PUTSTATIC));}
        il.append(factory.createReturn(Type.VOID));
        method.setMaxStack();
        method.setMaxLocals();
        jclg.addMethod(method.getMethod());
        il.dispose();} 
    
        /** extends an existing <clinit> method which binds the static getter-variables to the corresponding getter-methods. 
          * Example:  private static Method getter_getAddress = J2OSynchroniser.getMethod(TestPerson1.class,"getAddress");*/
      private void extendCLINIT(String className, MethodGen clinit, ArrayList<Method> getters, InstructionFactory factory, ConstantPoolGen cp) {
        InstructionList il = new InstructionList();
        LDC_W ldc = new LDC_W(cp.lookupClass(className));
        for(Method getter : getters) {
            String name = getter.getName();
            il.append(ldc);
            il.append(new PUSH(cp, name));
            il.append(factory.createInvoke("Java2OWL.J2OSynchroniser", "getMethod", new ObjectType("java.lang.reflect.Method"), new Type[] { new ObjectType("java.lang.Class"), Type.STRING }, Constants.INVOKESTATIC));
            il.append(factory.createFieldAccess(className, getterVariable(getter), new ObjectType("java.lang.reflect.Method"), Constants.PUTSTATIC));}
        InstructionList ilold = clinit.getInstructionList();
        ilold.insert(il);
        clinit.setMaxStack();
        clinit.setMaxLocals();
        il.dispose();}
    
    /** This method creates the test method for deciding whether synchronisation should be done. 
       * <p>Something like 
       * <pre>
       *  J2OIndividualManager J2OSynchronisationActive() {
       *    Java2OWLFlags |= 6;
       *    return  (Java2OWLFlags & 1) == 0 ? null : j2OWLIndividualManager;}
       </pre>
       * <p>The generated code is optimised to avoid double loading of Java2OWLFlags.
       * <p>Notice that calling this method for an object of class C uses Java2OWLFlags of this
       * class C, and not of a superclass of C, even if the method is called from within an inherited method.
       * This way the same variable Java2OWLFlags may have a different value for superclasses of C.
       * <p>The default of the Java2OWLFlags after object creation is 0. 
       * This should mean: synchronisation is active. <br/>
       * (Java2OWLFlags & 1) == 1 means synchronisation is not active!
       */
    
    private void createSyncTest(String className, ClassGen jclg, InstructionFactory factory,ConstantPoolGen cp) {
        InstructionList il = new InstructionList();
        ObjectType type = new ObjectType("Java2OWL.J2OIndividualManager");
        MethodGen method = new MethodGen(0, type, Type.NO_ARGS, new String[] {  }, "J2OSynchronisationActive", className, il, cp);
       
        InstructionHandle ih_0 = il.append(factory.createLoad(Type.OBJECT, 0));
        il.append(InstructionConstants.DUP);
        il.append(factory.createFieldAccess(className, SYNCHRONIZE_BYOBJECT_VARIABLE, Type.INT, Constants.GETFIELD));
        il.append(new PUSH(cp, 6));
        il.append(InstructionConstants.IOR);
        il.append(InstructionConstants.DUP_X1);
        il.append(factory.createFieldAccess(className, SYNCHRONIZE_BYOBJECT_VARIABLE, Type.INT, Constants.PUTFIELD));
        il.append(new PUSH(cp, 1));
        il.append(InstructionConstants.IAND);
        BranchInstruction ifeq_17 = factory.createBranchInstruction(Constants.IFEQ, null);
        il.append(ifeq_17);
        il.append(InstructionConstants.ACONST_NULL);
        BranchInstruction goto_21 = factory.createBranchInstruction(Constants.GOTO, null);
        il.append(goto_21);
        InstructionHandle ih_24 = il.append(factory.createFieldAccess(className, INDIVIDUAL_MANAGER_VARIABLE, new ObjectType("Java2OWL.J2OIndividualManager"), Constants.GETSTATIC));
        InstructionHandle ih_27 = il.append(factory.createReturn(Type.OBJECT));
        ifeq_17.setTarget(ih_24);
        goto_21.setTarget(ih_27);
        method.setMaxStack();
        method.setMaxLocals();
        jclg.addMethod(method.getMethod());
        il.dispose();}
    
    /** returns jcl's method with the given name. 
     * Exceptions are printed, but otherwise suppressed.
     */
    public static java.lang.reflect.Method getMethod(Class jcl, String name) {
        try {return jcl.getMethod(name); } catch (Exception ex) {ex.printStackTrace();}
        return null;
    }
    
    public static void main(String[] args) {
        
    }
} 

