 /*
    * Copyright (c) 2012, H. J. Ohlbach. All rights reserved.
    * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
    *
    * This code is free software; you can redistribute it and/or modify it
    * under the terms of the GNU General Public License version 2 only, as
    * published by the Free Software Foundation. 
    *
    * This code is distributed in the hope that it will be useful, but WITHOUT
    * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    * version 2 for more details.
    */

package Java2OWL;

import java.lang.reflect.*;
import java.util.*;
import java.util.Map.Entry;
import org.semanticweb.owlapi.model.*;

/** This class represents the correspondences between Java accessor groups and OWL-Properties. 
 * <p> This is an abstract superclass for several classes specific for accessor group types.
 * 
 *  <p>Methods in several Java class can be mapped to the same OWL-Property.
 *  An example is a "has-part"-property which may occur in different Java classes, 
 *  but is mapped to a single OWL-Property.
 *  Different methods in the same class, however, cannot be mapped to same OWL-property.
 *  Properties may be data properties and object properties. 
 * <p>
 * Typical application:<br/>
 * A new PropertyWrapper is created for the first occurrence of a getter-method to be mapped to OWL.
 * Further getter-methods in other classes to be mapped to the same OWL-property are added by the join-method.
 * As soon as all methods have been processed, getInconsistencies() can be called to check for 
 * internal inconsistencies. 
 * If there are no inconsistencies, createOWLAxioms can be called to get the corresponding OWL-Axioms for 
 * the OWL-Property.
 * <p>Notice that a PropertyWrapper need not analyse and store all
 * setter, adder and remover methods, because the J2OSynchroniser analyses the annotations 
 * separately. A PropertyWrapper needs only store the information which is necessary for 
 * the compilation of Java to OWL.
 *
 * @author ohlbach
 */

abstract public class PropertyWrapper {
    /** the top manager. */
    protected J2OManager manager; 
    
    /** for logging messages. */
    protected Logging logger = null;
    
    
    /** the OWL-name of the property. */
    protected String name;
    
    /** returns the name of the OWL-property. */
    String getName() {return name;}

    
    /** the corresponding OWL-property. */
    protected OWLProperty property = null;
    
    /** true if the property is from the background ontology. 
     *<p>Properties from the background ontology need not generate axioms.*/
    private boolean fromBackground = false;
    
    /** returns the OWL-property */
    OWLProperty getProperty() {return property;}
    
    /** sets the corresponding OWL-property */
    void setProperty(OWLProperty property, boolean fromBackground) {
        this.property = property;
        this.fromBackground = fromBackground;}
    
    
    /** maps the owner class to the component. 
     * <p>Different getter methods in different classes can be mapped to the same OWL proprty.
     * Therefore we need a data structure for each occurrence.*/
    protected Map<Class,PropertyWrapperComponent> pwComponents = new HashMap<Class,PropertyWrapperComponent>();
    
    /** returns the property wrapper component for the given class. */
    PropertyWrapperComponent getPropertyWrapperComponent(Class jcl) {return pwComponents.get(jcl);}
    
    /** returns all PropertyWrapperComponents */
    Map<Class,PropertyWrapperComponent> getPropertyWrapperComponents() {return pwComponents;}
    
    /** is only used during construction of join. */
    protected PropertyWrapperComponent currentPWC;
    
    /** is only used during construction of join. */
    protected Class currentClass;
   
    
    /** true if it is a data property. */
    boolean isDataProperty;
    
    /** indicates a functional property. */
    protected boolean isFunctional = false;
    
    /** returns true if the property is isFunctional */
    boolean isFunctional() {return isFunctional;}

    /** returns true if the property is a data property. */
    boolean isDataProperty() {return isDataProperty;}
    
    /** indicates that the property is a mapper. */
    boolean isMapper;
    
    /** returns true if the property is a mapper property. */
    boolean isMapper() {return isMapper;}
    
    protected boolean symmetric = false;
    protected boolean asymmetric = false;
    protected boolean reflexive = false;
    protected boolean irreflexive = false;
    protected boolean transitive = false;
    
    
    /** controls whether range axioms should be added to the OWL-axioms */
    protected boolean addRangeAxiom = false;
        
   /** for Mapping properties: maps keys to OWL properties */
    protected Map<Object,OWLProperty> key2Property = new HashMap<Object,OWLProperty>();  
    
    /** This constructor is only for test purposes! */
    PropertyWrapper(String name) {this.name = name;}
    
    /** constructs a new PropertyWrapper for the given getter.
     * 
     * <p>
     * The given Java class may be a subclass of the class where the getter is declared.
     * The information is taken from the getter itself and from the annotation 2OWLProperty.
     * Besides the getters-getter itself, setter-methods, adder-methods and remover-methods are 
     * identified and stored. 
     * 
     * @param jcl a Java-class
     * @param getter a getters-getter.
     * @throws J2OException if the return type of the getter cannot be mapped to an OWL-Class.
     */
    PropertyWrapper(Class jcl, Method getter, J2OManager manager) throws J2OException {
        this.manager = manager;
        this.logger = manager.getLogger();
        currentClass = jcl;
        logger.entering(jcl,getter);
        String errors = intialise(getter);
        logger.exiting(jcl,getter);
        if(errors != null) throw new J2OException(errors);}
    
    /** joins to an already existing PropertyWrapper a getter of another class. 
     * <p>
     * This method should be called if further getter-methods are mapped to the same OWL-property.
     * The new data is generated in a similar way as in the constructor.
     * Two methods in the same class can not be mapped to the same OWL-property.
     * 
     * @param jcl is the Java class with the getter method.
     * @param getter the getter-method. 
     * @return an error string if the method cannot be mapped to OWL.
     * @throws J2OException if something goes wrong.
     */
    void join(Class jcl, Method getter) throws J2OException {
        logger.entering(jcl,getter);
        PropertyWrapperComponent prwc = getPropertyWrapperComponent(jcl); 
        if(prwc != null) {
            String message = "Method " + getter.getName()+" is already mapped to " + name + " in class " + jcl.getName();
            logger.error(message, Logging.Origin.USAGE);
            logger.exiting(jcl,getter);
            throw new J2OException(message);}
        
        currentClass = jcl;
        String errors = intialise(getter);
        logger.exiting(jcl,getter);
        if(errors != null) {
            errors = "Errors in property " + getName() + " in class " + jcl.getName() + "\n" + errors;
            throw new J2OException(errors);}
        errors = getInconsistencies();
        if(errors == null) return;
        errors = "Inconsistencies in property " + getName() + " in class " + jcl.getName() + "\n" + errors;
        logger.error(errors,Logging.Origin.USAGE); 
        throw new J2OException(errors);
    }
    
    /** constructs all the internal data structures for the property derived from the getter. */
    private String intialise(Method getter) {
        currentPWC = new PropertyWrapperComponent(currentClass,getter);
        StringBuffer errors = new StringBuffer();
        StringBuffer warnings = new StringBuffer();
        if(this.determineClasses(errors)) {
            J2OWLProperty annotation = (J2OWLProperty)(getter.getAnnotation(J2OWLProperty.class));
            name = getPropertyName(getter); 
            addProperties(annotation,errors);
            findAccessors(annotation, errors, warnings);}
        
        if(errors.length() != 0) {
            String message = "Class " + currentClass.getCanonicalName() + ", Method " + getter.getName() + "\n";
            errors.insert(0,"\n");
            logger.error(message,Logging.Origin.USAGE,errors);
            return message + errors.toString();}
        
        if(warnings.length() != 0) {
            String message = "Class " + currentClass.getCanonicalName() + ", Method " + getter.getName() + "\n";
            warnings.insert(0,"\n");
            logger.warning(message,Logging.Origin.USAGE,warnings);}
        
        
        currentPWC.isDataProperty = Utilities.isOWLDatatype(currentPWC.rangeClass);
        pwComponents.put(currentClass,currentPWC); 
        return null;}
    
    
    /** Searches for a method of the given accessor type.
     * 
     * @param accessorType the type of the accessor.
     * @param annotation the annotation for the accessorType, a comma separated names of the accessor methods.
     * @param required if true then an accessor of the type is required. 
     * @param errors for collecting error messages.
     * @param warnings the collection of warnings. 
     * @return the accessor method or null if none could be found. 
     * <p> All accessor methods are type checked, but only one is returned.
     */
    protected Method findAccessor(String accessorType, String annotation, boolean required, StringBuffer errors, StringBuffer warnings) {
        if(annotation.isEmpty()) return null;
        Method accessorMethod = null;
        for(String name : annotation.split(",")) {
            name = name.trim(); 
            for(Method accessor : currentClass.getMethods()) { 
                if(name.equals(accessor.getName())){
                    accessor = checkAccessor(accessor,accessorType,errors,warnings);
                    if(accessor != null) accessorMethod = accessor;}}}
        if(required && accessorMethod == null) 
            warnings.append("No suitable " + accessorType + " found for " + name + "\n");
        return accessorMethod;}
    
    /** determines the key class and the range class of the getter method
     * 
     * @param errors collects error messages.
     * @return true if the classes are okay.
     */
    abstract protected boolean determineClasses(StringBuffer errors);
    
    /** finds all setter, getter, adder, remover, clearer, getKey methods. */
    abstract protected void findAccessors(J2OWLProperty annotation, StringBuffer errors, StringBuffer warnings);
    
    /** checks whether the accessor has the right argument types. */
    private Method checkAccessor(Method accessor, String accessorType, StringBuffer errors, StringBuffer warnings) {
        Class[][] types = currentPWC.types.get(accessorType);
        if(types == null) {warnings.append("Method " + accessor.getName() + " of type " + accessorType + " is ignored.\n"); return null;}
        Class[] argumentClasses = accessor.getParameterTypes();
        int arguments = argumentClasses.length;
        boolean exactLength = currentPWC.exactLength.get(accessorType);
        boolean correct = true;
        boolean found = false;
        for(int t = 0; t < types.length; ++t) {
            Class[] option = types[t];
            found = (t == 0) && (option.length == arguments);
            if(exactLength ? (option.length != arguments) : (option.length > arguments)) {correct = false; continue;}
            int length = option.length;
            for(int i = 0; i < length; ++i)  {
                if(!(argumentClasses[i].isAssignableFrom(option[i]) || Utilities.isBoxable(option[i], argumentClasses[i]))) {
                    correct = false; break;}}
            if(correct) break;
            correct = true;}
        if(!correct) {
            errors.append("The " + accessorType + " " + accessor.getName() + " has incompatible types: ");
            for(Class argument : argumentClasses) errors.append(argument.getName() + ",");
            errors.append("\nIt should have one of the following argument types: \n");
            for(Class[] option : types) 
                for(Class argument : option) errors.append(argument.getName() + ",");
            errors.append("\n");
            return null;}
        return found ? accessor : null;}
        
    
    
    /** adds the properties specified by the annotation to the property wrapper. 
     * 
     * @param annotation the getter annotation
     * @param errors for appending error messages.
     * @return true if there is no error.
     */
    final boolean addProperties(J2OWLProperty annotation, StringBuffer errors) {
        symmetric   |= annotation.symmetric();
        asymmetric  |= annotation.asymmetric();
        transitive  |= annotation.transitive();
        reflexive   |= annotation.reflexive();
        irreflexive |= annotation.irreflexive();
        addRangeAxiom |= annotation.addRangeAxiom();
        currentPWC.atleast  = annotation.atleast();
        currentPWC.atmost   = annotation.atmost();
        currentPWC.isTotal  = annotation.total();
        if(currentPWC.atmost < currentPWC.atleast) {
            errors.append("atmost < atleast: " + currentPWC.atmost + " < " + currentPWC.atleast);
            return false;}
        return true;}
            
    
    /** generates a name for the OWL-property. 
     * 
     * <p>
     * The name is either taken from the annotation, or from the method name. 
     * If the annotation contains local() = true then the name is prefixed with the class name.  
     * @param method the method
     * @return a name for the OWL-property. 
     */
    static String getPropertyName(Method method) {
        J2OWLProperty annotation = (J2OWLProperty)(method.getAnnotation(J2OWLProperty.class));
        String name = annotation.name();
        if(name.isEmpty()) name  = method.getName();
        if(annotation.local()) name = method.getDeclaringClass().getSimpleName()+"_"+name;
        return name;}
    
  
    
    /** returns an error-string if the data contains internal inconsistencies or null if not. 
     * <p>Side effect: the variables isFunctional isDataProperty, isMapper are set. */ 
    String getInconsistencies() {
        String message = "";
        if(symmetric && asymmetric)     message = "cannot be symmetric and asymmetric.\n";
        if(reflexive && irreflexive)    message += "cannot be reflexive and irreflexive.\n";
        Class keyClass = null;
        currentPWC = null;
        for(Entry<Class,PropertyWrapperComponent> pwc : pwComponents.entrySet()) {
            if(currentPWC == null) {
                currentPWC = pwc.getValue();
                isFunctional = currentPWC.isFunctional;
                isDataProperty = currentPWC.isDataProperty;
                isMapper = currentPWC.isMapper;
                keyClass = currentPWC.keyClass;}
            else {
                String jcl = pwc.getKey().getName();
                PropertyWrapperComponent pr = pwc.getValue();
                if(isFunctional != pr.isFunctional)     message += "can not be functional and relational in " + jcl +".\n";
                if(isDataProperty != pr.isDataProperty) message += "can not be data property and object property in " + jcl +".\n";
                if(isMapper != pr.isMapper)             message += "can not be a mapper an no mapper in " + jcl +".\n";
                if(keyClass != pr.keyClass)             message += "has different key classes in " + jcl +".\n";}}
        
        if(message.isEmpty()) return null;
        return "Inconsistencies in property " + name + "\n" + message;}
    
    Map<OWLProperty,Boolean> axiomsAdded = new HashMap<OWLProperty,Boolean>();
    ArrayList<OWLAxiom> axioms = new ArrayList<OWLAxiom>();
     
    /** generates OWL-Axioms for the concrete mapper property. */
    ArrayList<OWLAxiom> createOWLAxiomsForMappers(Class jcl, OWLProperty property) {
        J2OClassManager classManager = manager.getClassManager();
        axioms.clear();
        if(axiomsAdded.get(property) == null) {
            addGeneralOWLAxioms(axioms,property);
            axiomsAdded.put(property,true);}
        pwComponents.get(jcl).addOWLAxioms(axioms, property, classManager);
        return axioms;}
    
      /** generates OWL-Axioms for properties which are not mapper properties.
     * 
     * @param classManager contains the mapping of Java classes to OWL-classes. 
     * @return the list of OWL-axioms. 
     */
    ArrayList<OWLAxiom> createOWLAxioms() {
        J2OClassManager classManager = manager.getClassManager();
        axioms.clear();
        if(fromBackground || isMapper) return axioms;
        assert(property != null) : "The OWLProperty must be set.";
        addGeneralOWLAxioms(axioms,property);
        for(Entry<Class,PropertyWrapperComponent> entry: pwComponents.entrySet()) 
            entry.getValue().addOWLAxioms(axioms, property, classManager);
        return axioms;}
    
    /** generates the OWL-Axioms which are independent of the domain.
     * 
     * @param classManager contains the mapping of Java classes to OWL-classes. 
     * @return the list of OWL-axioms. 
     */
    private ArrayList<OWLAxiom> addGeneralOWLAxioms(ArrayList<OWLAxiom> axioms,OWLProperty property) {
        J2OClassManager classManager = manager.getClassManager();
        OWLDataFactory dataFactory = J2OOntologyManager.dataFactory;
        if(isDataProperty) { 
            OWLDataProperty pr = (OWLDataProperty)property;
            axioms.add(dataFactory.getOWLDeclarationAxiom(pr));
            if(addRangeAxiom) { // now we generate range axioms
                Set<OWLDataRange> odranges = new HashSet<OWLDataRange>();
                for(PropertyWrapperComponent pwrc : pwComponents.values()) 
                    odranges.add(J2OOntologyManager.toOWLDatatype.get(pwrc.rangeClass));
                if(!odranges.isEmpty()) 
                    axioms.add(dataFactory.getOWLDataPropertyRangeAxiom(pr, 
                            (odranges.size() == 1) ? odranges.iterator().next() :
                            dataFactory.getOWLDataUnionOf(odranges)));}
            if(isFunctional)  axioms.add(dataFactory.getOWLFunctionalDataPropertyAxiom(pr));}
        
        else { // the property is an object-property.
            OWLObjectProperty pr = (OWLObjectProperty)property; 
            axioms.add(dataFactory.getOWLDeclarationAxiom(pr));
            if(addRangeAxiom) {
                Set<OWLClassExpression> ocranges = new HashSet<OWLClassExpression>();
                for(PropertyWrapperComponent pwrc : pwComponents.values()) {
                    OWLClass ocl = classManager.getOWLClass(pwrc.rangeClass,false,false);
                    if(ocl != null) ocranges.add(ocl);}
                if(!ocranges.isEmpty()) {
                    axioms.add(dataFactory.getOWLObjectPropertyRangeAxiom(pr,
                            (ocranges.size() == 1) ? ocranges.iterator().next() :
                            dataFactory.getOWLObjectUnionOf(ocranges)));}}
            
            if(reflexive)    axioms.add(dataFactory.getOWLReflexiveObjectPropertyAxiom(pr));
            if(irreflexive)  axioms.add(dataFactory.getOWLIrreflexiveObjectPropertyAxiom(pr));
            if(symmetric)    axioms.add(dataFactory.getOWLSymmetricObjectPropertyAxiom(pr));
            if(asymmetric)   axioms.add(dataFactory.getOWLAsymmetricObjectPropertyAxiom(pr));   
            if(transitive)   axioms.add(dataFactory.getOWLTransitiveObjectPropertyAxiom(pr));         
            if(isFunctional)   axioms.add(dataFactory.getOWLFunctionalObjectPropertyAxiom(pr));}
        return axioms;}

   
    /** returns the getter method for the given class, or null if there is none. */ 
    Method getGetter(Class jcl) {
        PropertyWrapperComponent pwc = pwComponents.get(jcl);
        return (pwc == null) ? null : pwc.getter;}
    
    /** returns the setter method for the given class, or null if there is none. */ 
    Method getSetter(Class jcl) {
        PropertyWrapperComponent pwc = pwComponents.get(jcl);
        return (pwc == null) ? null : pwc.setter;}
    
    /** returns the adder method for the given class, or null if there is none. */ 
    Method getAdder(Class jcl) {
        PropertyWrapperComponent pwc = pwComponents.get(jcl);
        return (pwc == null) ? null : pwc.adder;}
    
    /** returns the adder method for the given class, or null if there is none. */ 
    Method getClearer(Class jcl) {
        PropertyWrapperComponent pwc = pwComponents.get(jcl);
        return (pwc == null) ? null : pwc.clearer;}
    
    /** returns the adder method for the given class, or null if there is none. */ 
    Method getGetKey(Class jcl) { 
        PropertyWrapperComponent pwc = pwComponents.get(jcl);
        return (pwc == null) ? null : pwc.getKey;}
    
     /** returns the range class for the given class. */ 
    Class getRangeClass(Class jcl) { 
        PropertyWrapperComponent pwc = pwComponents.get(jcl);
        return (pwc == null) ? null : pwc.rangeClass;}
    
    

    /** returns true if the property is asymmetric */
    boolean isAsymmetric() {return asymmetric;}

    /** returns true if the property is irreflexive */
    boolean isIrreflexive() {return irreflexive;}
 
    /** returns true if the property is reflexive */
    boolean isReflexive() {return reflexive;}

    /** returns true if the property is symmetric */
    boolean isSymmetric() {return symmetric;}

    /** returns true if the property is transitive */
    boolean isTransitive() {return transitive;}
    

    /** returns a description of the PropertyWrapper */
    @Override
    public String toString() {
        String s = this.getClass().getSimpleName() + " " + name;
        if(property != null) s += " OWL: "+ property.getIRI().toString();
        s += "\nProperties:\n";
        if(symmetric) s += "symmetric, ";
        if(asymmetric) s += "asymmetric, ";
        if(reflexive) s += "reflexive, ";
        if(irreflexive) s += "irreflexive, ";
        if(transitive) s += "transitive, ";
        if(addRangeAxiom) s += "add_range_axiom";
        s += "\n";
        for(PropertyWrapperComponent pwc :  pwComponents.values()) {
            s += pwc.toString() + "\n";}
  
        return s;}
        
        
    }
